import falcon
from .serialization import get_serializer


class BaseFalconResource:
    def __init__(self, resource):
        self.resource = resource

    def _check_method(self, method, action):
        if not self.resource.is_method_allowed(method, action):
            raise falcon.HTTPMethodNotAllowed(self.resource.get_allowed_methods(action))


class CollectionFalconResource(BaseFalconResource):
    """
    Objects collection base resource
    """

    def on_get(self, req, resp, **kwargs):
        """
        Get objects list
        :param req: request object
        :param resp: response object
        """
        self._check_method('get', 'list')

        result = self.resource.get_list(req, **kwargs)
        serializer = get_serializer(req)
        resp.body = serializer.dumps(result)

    def on_post(self, req, resp, **kwargs):
        """
        Create new object
        :param req: request object
        :param resp: response object
        """
        self._check_method('post', 'list')
        serializer = get_serializer(req)

        req.context['data'] = serializer.loads(req.stream.read().decode('utf-8'))
        result = self.resource.post_list(req, **kwargs)

        resp.status = falcon.HTTP_201
        resp.body = serializer.dumps(result)

    def on_put(self, req, resp, **kwargs):
        self._check_method('put', 'list')
        serializer = get_serializer(req)

        req.context['data'] = serializer.loads(req.stream.read().decode('utf-8'))

        results = self.resource.put_list(req, **kwargs)

        resp.status = falcon.HTTP_200
        resp.body = serializer.dumps(results)

    def on_delete(self, req, resp, **kwargs):
        self._check_method('delete', 'list')

        results = self.resource.delete_list(req, **kwargs)

        resp.status = falcon.HTTP_204
        serializer = get_serializer(req)
        resp.body = serializer.dumps(results)


class ItemFalconResource(BaseFalconResource):
    """
    Object item base resource
    """

    def on_get(self, req, resp, **kwargs):
        """
        Show object details
        :param pk: object id
        :param req: request object
        :param resp: response object
        """
        self._check_method('get', 'details')
        serializer = get_serializer(req)

        result = self.resource.get_details(req, **kwargs)

        resp.status = falcon.HTTP_200
        resp.body = serializer.dumps(result)

    def on_put(self, req, resp, **kwargs):
        """
        Update object
        :param pk: object id
        :param req: request object
        :param resp: response object
        """
        self._check_method('put', 'details')
        serializer = get_serializer(req)

        req.context['data'] = serializer.loads(req.stream.read().decode('utf-8'))
        result = self.resource.put_details(req, **kwargs)

        resp.status = falcon.HTTP_200
        resp.body = serializer.dumps(result)

    def on_delete(self, req, resp, **kwargs):
        """
        Delete object
        :param pk: object id
        :param req: request object
        :param resp: response object
        """
        self._check_method('delete', 'details')

        self.resource.delete_details(req, **kwargs)

        resp.status = falcon.HTTP_204


class FunctionResource():
    pass


def resource_func(func):
    def wrapper(req, resp, *args, **kwargs):
        result = func(req, resp, *args, **kwargs)

        if result is not None:
            serializer = get_serializer(req)
            resp.body = serializer.dumps(result)

    return wrapper


def create_func_resource(func, http_methods):
    resource = FunctionResource()

    for http_method in http_methods:
        method = 'on_' + http_method

        setattr(resource, method, resource_func(func))

    return resource



from datetime import datetime
from .errors import DoesNotExist


class Field(object):
    """
    Hydrator field

    Implement hydrate and dehydrate methods
    """

    def __init__(self, attribute, required=False, null=False, readonly=False, show=None):
        if show is None:
            show = ['details', 'list']

        self.attribute = attribute
        self.readonly = readonly
        self.required = required
        self.null = null
        self.show = show

    def hydrate(self, obj, value):
        """hydrate field to object"""
        if self.readonly:
            return

        value = self.convert(value)

        setattr(obj, self.attribute, value)

    def dehydrate(self, obj):
        """dehydrate field from object"""
        return getattr(obj, self.attribute)

    def convert(self, value):
        return value


class CharField(Field):
    pass


class TextField(Field):
    pass


class IntegerField(Field):
    def convert(self, value):
        return int(value)


class BooleanField(Field):
    def convert(self, value):
        return bool(value)


class DateTimeField(Field):
    def __init__(self, *args, format="%Y-%m-%d %H:%M:%S", **kwargs):
        super(DateTimeField, self).__init__(*args, **kwargs)
        self.format = format

    def convert(self, value):
        if value is None:
            return None

        return datetime.strptime(value, self.format)


class DateField(DateTimeField):
    def __init__(self, *args, format="%Y-%m-%d", **kwargs):
        super(DateField, self).__init__(*args, format=format, **kwargs)

    def convert(self, value):
        if value is None:
            return None

        return super(DateField, self).date()


class ForeignKeyField(Field):
    def __init__(self, attribute, rel_resource, *args, **kwargs):
        super(ForeignKeyField, self).__init__(attribute, *args, **kwargs)

        self.rel_resource = rel_resource

    def dehydrate(self, obj):
        """dehydrate field from object"""
        rel_obj = getattr(obj, self.attribute)

        return getattr(rel_obj, self.rel_resource._meta.primary_key)

    def convert(self, value):
        # todo: how to rewrite it normally?
        rel_pk = self.rel_resource._meta.primary_key
        rel_field = self.rel_resource.fields[rel_pk]
        filters = self.rel_resource._filters_from_kwargs(**{rel_pk: rel_field.convert(value)})

        qs = self.rel_resource._db.select(filters)[:1]

        if len(qs) == 0:
            raise DoesNotExist()

        return qs[0]


class ToManyField(Field):
    def __init__(self, attribute, rel_resource, full=False, *args, **kwargs):
        super(ToManyField, self).__init__(attribute, *args, **kwargs)

        self.rel_resource = rel_resource
        self.rel_pk = rel_resource._meta.primary_key
        self.full = full

    def dehydrate(self, obj):
        qs = getattr(obj, self.attribute)[:1000]

        return [self.dehydrate_rel_obj(rel_obj) for rel_obj in qs]

    def dehydrate_rel_obj(self, rel_obj):
        if self.full:
            return self.rel_resource.dehydrate(rel_obj, None)
        else:
            return getattr(rel_obj, self.rel_pk)

    def hydrate(self, obj, value):
        raise NotImplemented


from collections import defaultdict
from .falcon import ItemFalconResource, CollectionFalconResource, create_func_resource


class Api:
    def __init__(self, api_name):
        self.name = api_name
        self.resources = {}
        self.nested = defaultdict(list)

    def register_resource(self, resource):
        self.resources[resource._meta.resource_name] = resource

    def register_nested_resource(self, nested_resource_name, resource, foreign_key):
        if nested_resource_name not in self.resources:
            raise Exception()

        self.nested[nested_resource_name].append((resource, foreign_key))


def register_api(app, api):
    prefix = '/' + api.name
    for res in api.resources.values():
        register_resource(app, prefix, res)

    for res_name, nested_resources in api.nested.items():
        for nested_resource, foreign_key in nested_resources:
            register_resource(
                app,
                prefix + '/%s/{%s}' % (api.resources[res_name]._meta.resource_name, foreign_key),
                nested_resource
            )


def register_resource(app, prefix, resource):
    item_res = ItemFalconResource(resource)
    coll_res = CollectionFalconResource(resource)

    prefix += "/%s" % (resource._meta.resource_name)

    app.add_route(prefix + '/', coll_res)
    app.add_route(prefix + '/{%s}' % resource._meta.primary_key, item_res)

    for func_name, uri, http_methods in resource.custom_methods:
        func = getattr(resource, func_name)

        custom_res = create_func_resource(func, http_methods)
        app.add_route(prefix + uri, custom_res)



from .fields import Field
from .adapters import detect_adapter
from .errors import DoesNotExist, MissingRequiredField
import falcon
from .settings import DEFAULT_MAX_LIMIT, DEFAULT_LIMIT


def custom_method(uri, http_methods=None):
    if not http_methods:
        http_methods = ['get']

    if isinstance(http_methods, str):
        http_methods = [http_methods]

    def wrapper(func):
        func.type = 'custom'
        func.uri = uri
        func.http_methods = http_methods
        return func

    return wrapper


class ResourceOptions(object):
    allowed_methods = ['get', 'post', 'put', 'delete', 'patch']
    list_allowed_methods = None
    details_allowed_methods = None
    limit = DEFAULT_LIMIT
    max_limit = DEFAULT_MAX_LIMIT
    resource_name = None
    filtering = {}
    ordering = []
    object_class = None
    excludes = []
    primary_key = 'id'
    adapter = None

    def __new__(cls, meta=None):
        overrides = {}

        # Handle overrides.
        if meta:
            for override_name in dir(meta):
                # No internals please.
                if not override_name.startswith('_'):
                    overrides[override_name] = getattr(meta, override_name)

        allowed_methods = overrides.get('allowed_methods', ResourceOptions.allowed_methods)

        if overrides.get('list_allowed_methods', None) is None:
            overrides['list_allowed_methods'] = allowed_methods

        if overrides.get('details_allowed_methods', None) is None:
            overrides['details_allowed_methods'] = allowed_methods

        return object.__new__(type('ResourceOptions', (cls,), overrides))


class DeclarativeMetaclass(type):
    def __new__(cls, name, bases, attrs):
        new_class = super(DeclarativeMetaclass, cls).__new__(cls, name, bases, attrs)

        opts = getattr(new_class, 'Meta', None)
        new_class._meta = ResourceOptions(opts)

        new_class.fields = {}
        for field_name, obj in attrs.copy().items():
            if isinstance(obj, Field):
                new_class.fields[field_name] = attrs.pop(field_name)

        new_class.custom_methods = []
        for func_name, func in attrs.items():
            func_uri = getattr(func, 'uri', None)
            func_http_methods = getattr(func, 'http_methods', None)

            if getattr(func, 'type', None) is not None:
                new_class.custom_methods.append((func_name, func_uri, func_http_methods))

        return new_class


class ModelDeclarativeMetaclass(DeclarativeMetaclass):
    def __new__(cls, name, bases, attrs):
        new_class = super(ModelDeclarativeMetaclass, cls).__new__(cls, name, bases, attrs)

        if new_class._meta.object_class:
            if not new_class._meta.adapter:
                new_class._meta.adapter = detect_adapter(new_class._meta.object_class)

            new_class._db = new_class._meta.adapter(new_class._meta.object_class)

            new_class.fields.update(new_class._db.get_fields(new_class._meta.excludes + list(new_class.fields.keys())))

        return new_class


class Resource(metaclass=DeclarativeMetaclass):
    _meta = None

    fields = None

    def is_method_allowed(self, method, action):
        return method in self.get_allowed_methods(action)

    def get_allowed_methods(self, action):
        if action == 'list':
            return self._meta.list_allowed_methods
        elif action == 'details':
            return self._meta.details_allowed_methods
        else:
            return self._meta.allowed_methods

    def get_list(self, req, **kwargs):
        raise NotImplemented

    def post_list(self, req, **kwargs):
        raise NotImplemented

    def put_list(self, req, **kwargs):
        raise NotImplemented

    def delete_list(self, req, **kwargs):
        raise NotImplemented

    def get_details(self, req, **kwargs):
        raise NotImplemented

    def put_details(self, req, **kwargs):
        raise NotImplemented

    def delete_details(self, req, **kwargs):
        raise NotImplemented

    def dehydrate(self, obj, sender=None):
        """
        Dehydrate object
        :param obj: object
        """
        return {key: field.dehydrate(obj) for key, field in self.fields.items() if
                sender is None or (sender in field.show)}

    def hydrate(self, obj, data):
        """
        Hydrate data to object
        :param obj: object
        :param data: data as dictionary
        """
        for key, field in self.fields.items():
            if field.readonly:
                continue

            if key not in data:
                if field.required:
                    raise MissingRequiredField('Field "%s" is required' % key)

                continue

            value = data.get(key, None)

            field.hydrate(obj, value)

    def _filters_from_request(self, req):
        filters = []
        for key, v in req.params.items():
            try:
                field, op = key.rsplit('__', 1)
            except Exception:
                field, op = key, 'exact'

            if field in self._meta.filtering and op in self._meta.filtering[field]:
                filters.append((field, op, v))

        return filters

    def _filters_from_kwargs(self, **kwargs):
        return [(k, 'exact', self.fields[k].convert(v)) for k, v in kwargs.items()]


class ModelResource(Resource, metaclass=ModelDeclarativeMetaclass):
    _db = None

    def get_list(self, req, **kwargs):
        try:
            filters = self._filters_from_request(req) + self._filters_from_kwargs(**kwargs)
        except DoesNotExist:
            raise falcon.HTTPNotFound()

        offset = req.get_param_as_int('offset', min=0) or 0
        limit = req.get_param_as_int('limit', min=1, max=self._meta.max_limit) or self._meta.limit

        qs = self._db.select(filters)
        meta, qs = self._paginate(qs, offset, limit)

        return {
            'meta': meta,
            'objects': [self.dehydrate(obj, sender='list') for obj in qs]
        }

    def post_list(self, req, **kwargs):
        data = req.context['data']

        obj = self._db.create_object()

        try:
            self.hydrate(obj, data)
        except MissingRequiredField as e:
            raise falcon.HTTPBadRequest('Missing field', e.args[0])

        self._db.save_object(obj)

        return {
            self._meta.primary_key: getattr(obj, self._meta.primary_key)
        }

    def delete_list(self, req, **kwargs):
        try:
            filters = self._filters_from_request(req) + self._filters_from_kwargs(**kwargs)
        except DoesNotExist:
            raise falcon.HTTPNotFound()

        self._db.delete(filters)

    def put_list(self, req, **kwargs):
        data = req.context['data']

        pk_field = self._meta.primary_key

        results = []
        for item in data:
            pk = item.get(pk_field)

            if not pk:
                obj = self._db.create_object()
            else:
                try:
                    obj = self.find_object(**{pk_field: pk})
                except DoesNotExist:
                    raise falcon.HTTPBadRequest('', '')

            self.hydrate(obj, item)

            self._db.save_object(obj)

            results.append(self.dehydrate(obj, sender='list'))

        return {
            'objects': results
        }

    def get_details(self, req, **kwargs):
        try:
            obj = self.find_object(**kwargs)
        except DoesNotExist:
            raise falcon.HTTPNotFound()

        return self.dehydrate(obj, sender='details')

    def put_details(self, req, **kwargs):
        try:
            obj = self.find_object(**kwargs)
        except DoesNotExist:
            raise falcon.HTTPNotFound()

        data = req.context['data']

        self.hydrate(obj, data)
        self._db.save_object(obj)

        return self.dehydrate(obj, sender='details')

    def delete_details(self, req, **kwargs):
        affected_rows = self._db.delete(self._filters_from_kwargs(**kwargs))

        if not affected_rows:
            raise falcon.HTTPNotFound()

    def find_object(self, **kwargs):
        qs = self._db.select(self._filters_from_kwargs(**kwargs))[:1]

        if len(qs) == 0:
            raise DoesNotExist()

        return qs[0]

    def _paginate(self, qs, offset, limit):
        meta = {
            'limit': limit,
            'offset': offset,
            'total_count': qs.count(), # todo: make it using db adapter
        }

        return meta, qs[offset:offset + limit]



import peewee
from wing.fields import *


class Adapter(object):
    def __init__(self, cls):
        self.cls = cls

    def select(self, filters=None):
        query = self.cls.select()
        if filters:
            query = self.apply_filters(query, filters)

        return query

    def delete(self, filters=None):
        """
        Delete objects matching filters parameter
        :return count of deleted rows
        """
        query = self.cls.delete()

        if filters:
            query = self.apply_filters(query, filters)

        return query.execute()

    def create_object(self):
        return self.cls()

    def save_object(self, obj):
        return obj.save()

    def get_fields(self, excludes=None):
        fields = {}
        if not excludes: excludes = []

        for key, field in self.cls._meta.fields.items():
            if key not in excludes:
                field = create_resource_field(field)
                if field is not None:
                    fields[key] = field

        return fields

    def apply_filters(self, query, filters):
        conditions = [_filter_get_expr(getattr(self.cls, field), type, val) for field, type, val in filters if
                      hasattr(self.cls, field)]

        if conditions:
            return query.where(*conditions)

        return query


def _filter_get_expr(field, filter_type, value):
    if filter_type == 'exact':
        return field == value
    elif filter_type == 'gt':
        return field > value
    elif filter_type == 'gte':
        return field >= value
    elif filter_type == 'lt':
        return field < value
    elif filter_type == 'lte':
        return field <= value
    elif filter_type == 'contains':
        return field.contains(value)
    elif filter_type == 'startswith':
        return field.startswith(value)
    elif filter_type == 'endswith':
        return field.endswith(value)
    elif filter_type == 'is_null':
        return field.is_null(value)
    elif filter_type == 'in':
        value = value.split(',')
        return field.in_(value)
    else:
        raise NotImplemented

def create_resource_field(orm_field):
    """
    Create resource field based on ORM field type
    :param orm_field: ORM field
    """
    mapping = {
        peewee.PrimaryKeyField: IntegerField,
        peewee.CharField: CharField,
        peewee.TextField: TextField,
        peewee.IntegerField: IntegerField,
        peewee.BooleanField: BooleanField,
        peewee.DateTimeField: DateTimeField,
        peewee.DateField: DateField,
    }

    if isinstance(orm_field, peewee.Field):
        is_required = not orm_field.null and orm_field.default is None
        if isinstance(orm_field, peewee.PrimaryKeyField):
            is_required = False

        is_nullable = orm_field.null

        if isinstance(orm_field, peewee.ForeignKeyField):
            return

        cls = mapping.get(type(orm_field))

        if not cls:
            raise Exception('Unsupported field type %s' % type(orm_field))

        return cls(orm_field.name, is_required, null=is_nullable)


from datetime import datetime
import json


class JSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, datetime):
            return obj.strftime('%Y-%m-%d %H:%M:%S')
        else:
            return json.JSONEncoder.default(self, obj)


def dumps(obj):
    return json.dumps(obj, cls=JSONEncoder)


def loads(content):
    return json.loads(content)



from setuptools import setup

setup(
    name='falcon-wing',
    version='0.1.6',
    description='Falcon extension for building nice looking API',

    author='Vlad Bakin',

    install_requires=['falcon'],

    packages=[
        'wing',
        'wing.adapters',
        'wing.serialization',
    ]
)



from ..errors import NoAdapter


def detect_adapter(cls):
    try:
        import peewee

        if issubclass(cls, peewee.Model):
            from .peewee import Adapter
            return Adapter
    except ImportError:
        pass

    raise NoAdapter('No adapter found for class "{0}"'.format(cls.__name__))



from .. import settings


def get_serializer(req):
    type = settings.SERIALIZER_DEFAULT

    if settings.SERIALIZER_REQUEST_PARAM:
        type = req.get_param(settings.SERIALIZER_REQUEST_PARAM) or type

    return _get_serializer_by_type(type)

def _get_serializer_by_type(type):
    if type == 'json':
        from . import json

        return json
    elif type == 'yaml':
        from . import yaml

        return yaml

    raise NotImplemented